import asyncio
import datetime
import logging
import os
import sys

from random import choice

import aiohttp.web as aiohttp_web

from . import _root_handler

from wslink.launcher import (
    SessionManager,
    ProxyMappingManagerTXT,
    ProcessManager,
    validateKeySet,
    STATUS_BAD_REQUEST,
    STATUS_SERVICE_UNAVAILABLE,
    filterResponse,
    STATUS_OK,
    extractSessionId,
    STATUS_NOT_FOUND,
)

logger = logging.getLogger(__name__)

# ===========================================================================
# Launcher ENV configuration
# ===========================================================================

ENABLE_GET = int(os.environ.get("WSLINK_LAUNCHER_GET", 0))
ENABLE_DELETE = int(os.environ.get("WSLINK_LAUNCHER_DELETE", 0))


# ===========================================================================
# Class to implement requests to POST, GET and DELETE methods
# ===========================================================================


class LauncherResource(object):
    def __init__(self, options, config):
        self._options = options
        self._config = config
        self.time_to_wait = int(config["configuration"]["timeout"])
        self.field_filter = config["configuration"]["fields"]
        self.session_manager = SessionManager(
            config, ProxyMappingManagerTXT(config["configuration"]["proxy_file"])
        )
        self.process_manager = ProcessManager(config)

    def __del__(self):
        try:
            # causes an exception when server is killed with Ctrl-C
            logger.warning("Server factory shutting down. Stopping all processes")
        except:
            pass

    # ========================================================================
    # Handle POST request
    # ========================================================================

    async def handle_post(self, request):
        payload = await request.json()

        # Make sure the request has all the expected keys
        if not validateKeySet(payload, ["application"], "Launch request"):
            return aiohttp_web.json_response(
                {"error": "The request is not complete"}, status=STATUS_BAD_REQUEST
            )

        # Try to free any available resource
        id_to_free = self.process_manager.listEndedProcess()
        for id in id_to_free:
            self.session_manager.deleteSession(id)
            self.process_manager.stopProcess(id)

        # Create new session
        session = self.session_manager.createSession(payload)

        # No resource available
        if not session:
            return aiohttp_web.json_response(
                {"error": "All the resources are currently taken"},
                status=STATUS_SERVICE_UNAVAILABLE,
            )

        # Start process
        proc = self.process_manager.startProcess(session)

        if not proc:
            err_msg = "The process did not properly start. %s" % str(session["cmd"])
            return aiohttp_web.json_response(
                {"error": err_msg}, status=STATUS_SERVICE_UNAVAILABLE
            )

        return await self._waitForReady(session, request)

    # ========================================================================
    # Wait for session to be ready
    # ========================================================================

    async def _waitForReady(self, session, request):
        start_time = datetime.datetime.now()
        check_line = "ready_line" in self._config["apps"][session["application"]]
        count = 0

        while True:
            if self.process_manager.isReady(session, count):
                filterkeys = self.field_filter
                if session["secret"] in session["cmd"]:
                    filterkeys = self.field_filter + ["secret"]
                return aiohttp_web.json_response(
                    filterResponse(session, filterkeys), status=STATUS_OK
                )

            elapsed_time = datetime.datetime.now() - start_time

            if elapsed_time.total_seconds() > self.time_to_wait:
                # Timeout is expired, if the process is not ready now, mark the
                # session as timed out, clean up the process, and return an error
                # response
                session["startTimedOut"] = True
                self.session_manager.deleteSession(session["id"])
                self.process_manager.stopProcess(session["id"])

                return aiohttp_web.json_response(
                    {
                        "error": "Session did not start before timeout expired. Check session logs."
                    },
                    status=STATUS_SERVICE_UNAVAILABLE,
                )

            await asyncio.sleep(1)
            count += 1

    # =========================================================================
    # Handle GET request
    # =========================================================================

    async def handle_get(self, request):
        id = extractSessionId(request)

        if not id:
            message = "id not provided in GET request"
            logger.error(message)
            return aiohttp_web.json_response(
                {"error": message}, status=STATUS_BAD_REQUEST
            )

        logger.info("GET request received for id: %s" % id)

        session = self.session_manager.getSession(id)
        if not session:
            message = "No session with id: %s" % id
            logger.error(message)
            return aiohttp_web.json_response(
                {"error": message}, status=STATUS_BAD_REQUEST
            )

        # Return session meta-data
        return aiohttp_web.json_response(
            filterResponse(session, self.field_filter), status=STATUS_OK
        )

    # =========================================================================
    # Handle DELETE request
    # =========================================================================

    async def handle_delete(self, request):
        id = extractSessionId(request)

        if not id:
            message = "id not provided in DELETE request"
            logger.error(message)
            return aiohttp_web.json_response(
                {"error": message}, status=STATUS_BAD_REQUEST
            )

        logger.info("DELETE request received for id: %s" % id)

        session = self.session_manager.getSession(id)
        if not session:
            message = "No session with id: %s" % id
            logger.error(message)
            return aiohttp_web.json_response(
                {"error": message}, status=STATUS_NOT_FOUND
            )

        # Remove session
        self.session_manager.deleteSession(id)
        self.process_manager.stopProcess(id)

        message = "Deleted session with id: %s" % id
        logger.info(message)

        return aiohttp_web.json_response(session, status=STATUS_OK)


# =============================================================================
# Start the web server
# =============================================================================


def startWebServer(options, config):
    # import pdb; pdb.set_trace()
    # Extract properties from config
    log_dir = str(config["configuration"]["log_dir"])
    content = str(config["configuration"]["content"])
    endpoint = str(config["configuration"]["endpoint"])
    host = str(config["configuration"]["host"])
    port = int(config["configuration"]["port"])
    sanitize = config["configuration"]["sanitize"]

    # Setup logging
    logFileName = log_dir + os.sep + "launcherLog.log"
    formatting = "%(asctime)s:%(levelname)s:%(name)s:%(message)s"
    # create file handler which logs even debug messages
    fh = logging.FileHandler(logFileName, mode="w")
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(logging.Formatter(formatting))
    logging.getLogger("wslink").addHandler(fh)
    if options.debug:
        console = logging.StreamHandler(sys.stdout)
        console.setLevel(logging.INFO)
        formatter = logging.Formatter(formatting)
        console.setFormatter(formatter)
        logging.getLogger("wslink").addHandler(console)

    web_app = aiohttp_web.Application()

    launcher_resource = LauncherResource(options, config)

    if not endpoint.startswith("/"):
        endpoint = "/{0}/".format(endpoint)

    routes = [
        aiohttp_web.post(endpoint, launcher_resource.handle_post),
    ]

    if ENABLE_GET:
        routes.append(aiohttp_web.get(endpoint + "{id}", launcher_resource.handle_get))

    if ENABLE_DELETE:
        routes.append(
            aiohttp_web.delete(endpoint + "{id}", launcher_resource.handle_delete)
        )

    web_app.add_routes(routes)

    if len(content) > 0:
        web_app.router.add_route("GET", "/", _root_handler)
        web_app.add_routes([aiohttp_web.static("/", content)])

    aiohttp_web.run_app(web_app, host=host, port=port)



# -----------------------------------------------------------------------------
# DISCLAIMER
# -----------------------------------------------------------------------------
# This implementation is not full featured but just aim to showcase
# the generic backend in a way that it can be used by anything.
# For real integration, using inheritance like for aiohttp is recommended.
# -----------------------------------------------------------------------------
import tornado
import tempfile

from tornado.websocket import WebSocketHandler


class GenericMessage:
    def __init__(self, data):
        self.data = data


class WsLinkWebSocket(WebSocketHandler):
    def __init__(self, application, request, generic_server=None, **kwargs):
        super().__init__(application, request, **kwargs)
        self._ws = generic_server[generic_server.ws_endpoints[0]]
        self._me = None

    async def on_msg_from_generic(self, binary, content):
        self.write_message(content, binary)

    def open(self):
        self._me = self._ws.connect()
        self._me.on_message(self.on_msg_from_generic)

    async def on_message(self, message):
        is_binary = not isinstance(message, str)
        await self._me.send(is_binary, GenericMessage(message))

    def on_close(self):
        self._me.close()


def startWebServer(*args, **kwargs):
    raise NotImplementedError("Generic backend does not provide a launcher")


def create_webserver(server_config):
    # Reverse connection
    if "reverse_url" in server_config:
        raise NotImplementedError("Generic backend does not support reverse_url")

    from ..generic.core import create_webserver as create_generic_webserver

    generic_server = create_generic_webserver(server_config)

    # Handle static content
    www = tempfile.mkdtemp()
    generic_server.write_static_content(www)

    # Tornado specific
    handlers = [
        (r"/ws", WsLinkWebSocket, {"generic_server": generic_server}),
        (r"/(.*)", tornado.web.StaticFileHandler, {"path": www}),
    ]

    application = tornado.web.Application(handlers)
    application.listen(generic_server.port)

    return generic_server



import asyncio
import copy
import inspect
import logging
import msgpack
import os
import traceback

from wslink import schedule_coroutine
from wslink.publish import PublishManager
from wslink.chunking import generate_chunks, UnChunker

# from http://www.jsonrpc.org/specification, section 5.1
METHOD_NOT_FOUND = -32601
AUTHENTICATION_ERROR = -32000
EXCEPTION_ERROR = -32001
RESULT_SERIALIZE_ERROR = -32002
# used in client JS code:
CLIENT_ERROR = -32099

# 4MB is the default inside aiohttp
MAX_MSG_SIZE = int(os.environ.get("WSLINK_MAX_MSG_SIZE", 4194304))

logger = logging.getLogger(__name__)


class AbstractWebApp:
    def __init__(self, server_config):
        self._last_active_client_id = None
        self._config = server_config
        self._shutdown_task = None
        self._completion = asyncio.get_event_loop().create_future()
        self._app = None

    # -------------------------------------------------------------------------
    # Config helper
    # -------------------------------------------------------------------------

    @property
    def config(self):
        return self._config

    @property
    def timeout(self):
        return int(self.config.get("timeout", "0"))

    @property
    def host(self):
        return self.config.get("host", "127.0.0.1")

    @property
    def port(self):
        return int(self.config.get("port", "8080"))

    @property
    def handle_signals(self):
        return self.config.get("handle_signals", True)

    @property
    def ssl_context(self):
        return self.config.get("ssl", None)

    # -------------------------------------------------------------------------
    # In flight state
    # -------------------------------------------------------------------------

    @property
    def last_active_client_id(self):
        return self._last_active_client_id

    @last_active_client_id.setter
    def last_active_client_id(self, value):
        self._last_active_client_id = value

    # -------------------------------------------------------------------------
    # Implementation server class
    # -------------------------------------------------------------------------

    def set_app(self, app):
        self._app = app

    def get_app(self):
        return self._app

    @property
    def app(self):
        return self._app

    # -------------------------------------------------------------------------
    # Legacy / deprecated
    # -------------------------------------------------------------------------

    def get_config(self):
        print("DEPRECATED: get_config() use property instead")
        return self.config

    def set_config(self, config):
        print("DEPRECATED: set_config() use constructor instead")

    def get_last_active_client_id(self):
        print(
            "DEPRECATED: get_last_active_client_id() should be replaced by last_active_client_id"
        )
        return self.last_active_client_id

    # -------------------------------------------------------------------------
    # Life cycles
    # -------------------------------------------------------------------------

    def shutdown_schedule(self):
        self._shutdown_task = (
            schedule_coroutine(self.timeout, self.stop) if self.timeout > 0 else None
        )

    def shutdown_cancel(self):
        if self._shutdown_task is not None:
            logger.info("Canceling shutdown task")
            self._shutdown_task.cancel()
            self._shutdown_task = None

    # -------------------------------------------------------------------------
    # Server status
    # -------------------------------------------------------------------------

    @property
    def completion(self):
        return self._completion

    def get_port(self):
        return 0

    # -------------------------------------------------------------------------
    # Need override
    # -------------------------------------------------------------------------

    async def start(self, port_callback=None):
        pass

    async def stop(self):
        pass


class WslinkHandler(object):
    def __init__(self, protocol=None, web_app=None):
        self.serverProtocol = protocol
        self.web_app = web_app
        self.functionMap = {}
        self.attachmentsReceived = {}
        self.attachmentsRecvQueue = []
        self.connections = {}
        self.authentified_client_ids = set()
        self.attachment_atomic = asyncio.Lock()
        self.pub_manager = PublishManager()
        self.unchunkers = {}

        # Build the rpc method dictionary, assuming we were given a serverprotocol
        if self.getServerProtocol():
            protocolList = self.getServerProtocol().getLinkProtocols()
            protocolList.append(self.getServerProtocol())
            for protocolObject in protocolList:
                protocolObject.init(
                    self.publish,
                    self.addAttachment,
                    lambda: schedule_coroutine(0, self.web_app.stop),
                )
                test = lambda x: inspect.ismethod(x) or inspect.isfunction(x)
                for k in inspect.getmembers(protocolObject.__class__, test):
                    proc = k[1]
                    if "_wslinkuris" in proc.__dict__:
                        uri_info = proc.__dict__["_wslinkuris"][0]
                        if "uri" in uri_info:
                            uri = uri_info["uri"]
                            self.functionMap[uri] = (protocolObject, proc)
            self.pub_manager.registerProtocol(self)

    def setServerProtocol(self, protocol):
        self.serverProtocol = protocol

    def getServerProtocol(self):
        return self.serverProtocol

    @property
    def publishManager(self):
        return self.pub_manager

    @property
    def reverse_connection_client_id(self):
        return "reverse_connection_client_id"

    async def onConnect(self, request, client_id):
        self.unchunkers[client_id] = UnChunker()

        if not self.serverProtocol:
            return
        if hasattr(self.serverProtocol, "onConnect"):
            self.serverProtocol.onConnect(request, client_id)
        for linkProtocol in self.serverProtocol.getLinkProtocols():
            if hasattr(linkProtocol, "onConnect"):
                linkProtocol.onConnect(request, client_id)

    async def onClose(self, client_id):
        del self.unchunkers[client_id]

        if not self.serverProtocol:
            return
        if hasattr(self.serverProtocol, "onClose"):
            self.serverProtocol.onClose(client_id)
        for linkProtocol in self.serverProtocol.getLinkProtocols():
            if hasattr(linkProtocol, "onClose"):
                linkProtocol.onClose(client_id)

    async def handleSystemMessage(self, rpcid, methodName, args, client_id):
        rpcList = rpcid.split(":")
        if rpcList[0] == "system":
            if methodName == "wslink.hello":
                if (
                    args
                    and args[0]
                    and (type(args[0]) is dict)
                    and ("secret" in args[0])
                    and await self.validateToken(args[0]["secret"], client_id)
                ):
                    self.authentified_client_ids.add(client_id)
                    # Once a client is authenticated let the unchunker allocate memory unrestricted
                    self.unchunkers[client_id].set_max_message_size(
                        4 * 1024 * 1024 * 1024
                    )  # 4GB
                    await self.sendWrappedMessage(
                        rpcid,
                        {
                            "clientID": "c{0}".format(client_id),
                            "maxMsgSize": MAX_MSG_SIZE,
                        },
                        client_id=client_id,
                    )
                else:
                    await self.sendWrappedError(
                        rpcid,
                        AUTHENTICATION_ERROR,
                        "Authentication failed",
                        client_id=client_id,
                    )
            else:
                await self.sendWrappedError(
                    rpcid,
                    METHOD_NOT_FOUND,
                    "Unknown system method called",
                    client_id=client_id,
                )
            return True
        return False

    async def onMessage(self, is_binary, msg, client_id):
        if not is_binary:
            return

        full_message = self.unchunkers[client_id].process_chunk(msg.data)
        if full_message is not None:
            await self.onCompleteMessage(full_message, client_id)

    async def onCompleteMessage(self, rpc, client_id):
        logger.debug("wslink incoming msg %s", self.payloadWithSecretStripped(rpc))
        if "id" not in rpc:
            return

        # TODO validate
        # version = rpc["wslink"]
        rpcid = rpc["id"]
        methodName = rpc["method"]

        args = []
        kwargs = {}
        if ("args" in rpc) and isinstance(rpc["args"], list):
            args = rpc["args"]
        if ("kwargs" in rpc) and isinstance(rpc["kwargs"], dict):
            kwargs = rpc["kwargs"]

        # Check for system messages, like hello
        if await self.handleSystemMessage(rpcid, methodName, args, client_id):
            return

        # Prevent any further processing if token is not valid
        if not self.isClientAuthenticated(client_id):
            await self.sendWrappedError(
                rpcid,
                AUTHENTICATION_ERROR,
                "Unauthorized: Skip message processing",
                client_id=client_id,
            )
            return

        # No matching method found
        if not methodName in self.functionMap:
            await self.sendWrappedError(
                rpcid,
                METHOD_NOT_FOUND,
                "Unregistered method called",
                methodName,
                client_id=client_id,
            )
            return

        obj, func = self.functionMap[methodName]
        args.insert(0, obj)

        try:
            self.web_app.last_active_client_id = client_id
            results = func(*args, **kwargs)
            if inspect.isawaitable(results):
                results = await results

            if self.connections[client_id].closed:
                # Connection was closed during RPC call.
                return

            await self.sendWrappedMessage(
                rpcid, results, method=methodName, client_id=client_id
            )
        except Exception as e_inst:
            captured_trace = traceback.format_exc()
            logger.error("Exception raised")
            logger.error(repr(e_inst))
            logger.error(captured_trace)
            await self.sendWrappedError(
                rpcid,
                EXCEPTION_ERROR,
                "Exception raised",
                {
                    "method": methodName,
                    "exception": repr(e_inst),
                    "trace": captured_trace,
                },
                client_id=client_id,
            )

    def payloadWithSecretStripped(self, payload):
        payload = copy.deepcopy(payload)
        if "args" in payload:
            for arg in payload["args"]:
                if type(arg) is dict and "secret" in arg:
                    arg["secret"] = "*****"
        return payload

    async def validateToken(self, token, client_id):
        if not self.serverProtocol:
            return True
        token_tested = False
        if hasattr(self.serverProtocol, "validateToken"):
            token_tested = True
            if not await self.serverProtocol.validateToken(token, client_id):
                return False
        for linkProtocol in self.serverProtocol.getLinkProtocols():
            if hasattr(linkProtocol, "validateToken"):
                token_tested = True
                if not await linkProtocol.validateToken(token, client_id):
                    return False
        if token_tested:
            return True
        return token == self.serverProtocol.secret

    def isClientAuthenticated(self, client_id):
        return client_id in self.authentified_client_ids

    def getAuthenticatedWebsockets(self, client_id=None, skip_last_active_client=False):
        if skip_last_active_client:
            last_c = self.web_app.last_active_client_id
            return [
                self.connections[c]
                for c in self.connections
                if self.isClientAuthenticated(c) and c != last_c
            ]

        if client_id:
            if self.isClientAuthenticated(client_id):
                return [self.connections.get(client_id)]
            else:
                return []

        return [
            self.connections[c]
            for c in self.connections
            if self.isClientAuthenticated(c)
        ]

    async def sendWrappedMessage(
        self, rpcid, content, method="", client_id=None, skip_last_active_client=False
    ):
        wrapper = {
            "wslink": "1.0",
            "id": rpcid,
            "result": content,
        }

        try:
            packed_wrapper = msgpack.packb(wrapper)
        except Exception:
            # the content which is not serializable might be arbitrarily large, don't include.
            # repr(content) would do that...
            await self.sendWrappedError(
                rpcid,
                RESULT_SERIALIZE_ERROR,
                "Method result cannot be serialized",
                method,
                client_id=client_id,
            )
            return

        websockets = self.getAuthenticatedWebsockets(client_id, skip_last_active_client)

        # aiohttp can not handle pending ws.send_bytes()
        # tried with semaphore but got exception with >1
        # https://github.com/aio-libs/aiohttp/issues/2934
        async with self.attachment_atomic:
            for chunk in generate_chunks(packed_wrapper, MAX_MSG_SIZE):
                for ws in websockets:
                    if ws is not None:
                        await ws.send_bytes(chunk)

    async def sendWrappedError(self, rpcid, code, message, data=None, client_id=None):
        wrapper = {
            "wslink": "1.0",
            "id": rpcid,
            "error": {
                "code": code,
                "message": message,
            },
        }
        if data:
            wrapper["error"]["data"] = data

        try:
            packed_wrapper = msgpack.packb(wrapper)
        except Exception:
            del wrapper["error"]["data"]
            packed_wrapper = msgpack.packb(wrapper)

        websockets = (
            [self.connections[client_id]]
            if client_id
            else [self.connections[c] for c in self.connections]
        )
        # aiohttp can not handle pending ws.send_bytes()
        # tried with semaphore but got exception with >1
        # https://github.com/aio-libs/aiohttp/issues/2934
        async with self.attachment_atomic:
            for chunk in generate_chunks(packed_wrapper, MAX_MSG_SIZE):
                for ws in websockets:
                    if ws is not None:
                        await ws.send_bytes(chunk)

    def publish(self, topic, data, client_id=None, skip_last_active_client=False):
        client_list = [client_id] if client_id else [c_id for c_id in self.connections]
        for client in client_list:
            if self.isClientAuthenticated(client):
                self.pub_manager.publish(
                    topic,
                    data,
                    client_id=client,
                    skip_last_active_client=skip_last_active_client,
                )

    def addAttachment(self, payload):
        return self.pub_manager.addAttachment(payload)

    def setSecret(self, newSecret):
        self.secret = newSecret



import argparse
import asyncio
import os
import logging
import aiohttp
from aiohttp import web

# -----------------------------------------------------------------------------
# 4MB is the default inside aiohttp
# -----------------------------------------------------------------------------

MAX_MSG_SIZE = int(os.environ.get("WSLINK_MAX_MSG_SIZE", 4194304))
HEART_BEAT = int(os.environ.get("WSLINK_HEART_BEAT", 30))  # 30 seconds

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# -----------------------------------------------------------------------------
# Helper classes
# -----------------------------------------------------------------------------


class WsClientConnection:
    def __init__(self, propagate_disconnect=True):
        self._url = None
        self._session = None
        self._ws = None
        self._connected = 0
        self._destination = None
        self._ready = asyncio.get_running_loop().create_future()
        self.propagate_disconnect = propagate_disconnect

    def bind(self, value):
        self._destination = value

    @property
    def ready(self):
        return self._ready

    async def connect(self, url):
        logger.debug("client::connect::%s", url)
        self._url = url
        if self._session is None:
            async with aiohttp.ClientSession() as session:
                logger.debug("client::connect::session")
                self._session = session
                try:
                    async with session.ws_connect(self._url) as ws:
                        logger.debug("client::connect::ws")
                        self._ws = ws
                        self._connected += 1
                        self._ready.set_result(True)
                        async for msg in ws:
                            logger.debug("client::connect::ws::msg")
                            if self._connected < 1:
                                logger.debug("client::connect::ws::msg::disconnect")
                                break

                            if self._destination:
                                logger.debug("client::connect::ws::msg::send")
                                await self._destination.send(msg)
                            else:
                                logger.error("ws-client: No destination for message")

                        # Disconnect
                        self.disconnect()

                        # Cleanup connection
                        if not self._ws.closed:
                            await self._ws.close()
                        self._ws = None
                        self._session = None
                finally:
                    self._ready.set_result(False)

        logger.debug("client::connect::exit")

    async def send(self, msg):
        if self._connected > 0 and not self._ws.closed:
            logger.debug("client::send")
            if msg.type == aiohttp.WSMsgType.TEXT:
                await self._ws.send_str(msg.data)
            elif msg.type == aiohttp.WSMsgType.BINARY:
                await self._ws.send_bytes(msg.data)
            elif msg.type == aiohttp.WSMsgType.PING:
                await self._ws.ping(msg.data)
            elif msg.type == aiohttp.WSMsgType.PONG:
                await self._ws.pong(msg.data)
            elif msg.type == aiohttp.WSMsgType.CLOSE:
                await self._ws.close()
            else:
                logger.error("Invalid message to forward")
        else:
            logger.error("client::send - NO SEND")
            logger.error("%s - %s", self._connected, self._ws.closed)
            logger.error("-" * 60)

    def disconnect(self):
        logger.debug("client::disconnect %s", self._connected)
        if self._connected > 0:
            self._connected = -1
            if self._destination and self.propagate_disconnect:
                self._destination.disconnect()

    async def close(self):
        if self._ws is not None:
            await self._ws.close()


# -----------------------------------------------------------------------------


class WsServerConnection:
    def __init__(self, propagate_disconnect=True):
        self._ws = None
        self._destination = None
        self._connected = 0
        self.propagate_disconnect = propagate_disconnect

    def bind(self, value):
        self._destination = value

    async def connect(self, request):
        logger.debug("server::connect")
        self._ws = web.WebSocketResponse(
            max_msg_size=MAX_MSG_SIZE, heartbeat=HEART_BEAT
        )
        await self._ws.prepare(request)
        logger.debug("server::connect::prepare")
        self._connected += 1

        if self._connected > 0:
            async for msg in self._ws:
                logger.debug("server::connect::ws::msg")
                if self._connected < 1:
                    break

                if self._destination:
                    logger.debug("server::connect::ws::msg::send-begin")
                    await self._destination.send(msg)
                    logger.debug("server::connect::ws::msg::send-end")
                else:
                    logger.error("ws-server: No destination for message")

        # Disconnect
        self.disconnect()

        # Cleanup connection
        if not self._ws.closed:
            await self._ws.close()
        self._ws = None

        logger.debug("server::connect::exit")

    async def send(self, msg):
        if self._connected > 0 and not self._ws.closed:
            logger.debug("server::send")
            if msg.type == aiohttp.WSMsgType.TEXT:
                await self._ws.send_str(msg.data)
            elif msg.type == aiohttp.WSMsgType.BINARY:
                await self._ws.send_bytes(msg.data)
            elif msg.type == aiohttp.WSMsgType.PING:
                await self._ws.ping(msg.data)
            elif msg.type == aiohttp.WSMsgType.PONG:
                await self._ws.pong(msg.data)
            elif msg.type == aiohttp.WSMsgType.CLOSE:
                await self._ws.close()
            else:
                logger.error("Invalid message to forward")
        else:
            logger.error("server::send - NO SEND")
            logger.error("%s - %s", self._connected, self._ws.closed)
            logger.error("-" * 60)

    def disconnect(self):
        logger.debug("server::disconnect %s", self._connected)
        if self._connected > 0:
            self._connected = -1
            if self._destination and self.propagate_disconnect:
                self._destination.disconnect()

    async def close(self):
        if self._ws is not None:
            await self._ws.close()


# -----------------------------------------------------------------------------


class ForwardConnection:
    def __init__(self, request, url):
        self._req = request
        self._url = url
        self._ws_client = WsClientConnection()
        self._ws_server = WsServerConnection()
        self._ws_server.bind(self._ws_client)
        self._ws_client.bind(self._ws_server)

    async def connect(self):
        task = asyncio.create_task(self._ws_client.connect(self._url))
        task.add_done_callback(lambda *args, **kwargs: self._ws_server.disconnect())
        await self._ws_client.ready
        await self._ws_server.connect(self._req)

    def disconnect(self):
        self._ws_client.disconnect()
        self._ws_server.disconnect()


class SinkConnection:
    def __init__(self, request):
        self._process_req = request
        self._client_req = None
        self._process_ws = None
        self._client_ws = None

    def can_handle(self, request):
        if self._process_req == request:
            return True

        if self._client_req == request:
            return True

        if self._client_ws is None:
            return True

        return False

    async def connect(self, request):
        if self._process_req == request:
            # First connection is the actual server. Cannot reconnect.
            self._process_ws = WsServerConnection()
            await self._process_ws.connect(request)
            if self._client_ws is not None:
                await self._client_ws.close()

            return True
        elif self._client_req is None:
            # Second connection is the browser. Can reconnect.
            self._client_req = request
            self._client_ws = WsServerConnection(propagate_disconnect=False)
            self._client_ws.bind(self._process_ws)
            self._process_ws.bind(self._client_ws)
            await self._client_ws.connect(request)

            self._client_ws.bind(None)
            self._process_ws.bind(None)

            self._client_req = None
            self._client_ws = None

        return False


# -----------------------------------------------------------------------------
# Handlers
# -----------------------------------------------------------------------------


async def _root_handler(request):
    if request.query_string:
        return web.HTTPFound(f"index.html?{request.query_string}")
    return web.HTTPFound("index.html")


# -----------------------------------------------------------------------------


class WsHandler:
    def __init__(self):
        self._forward_map = {}
        self._relay_map = {}

    def get_handler(self, mode="forward"):
        logger.info("get_handler %s", mode)
        if mode == "forward":
            return self.forward_connect
        if mode == "relay":
            return self.relay_connect
        logger.error("No handler !!!")

    # -----------------------------
    # forward infrastructure
    # -----------------------------

    async def forward_connect(self, request):
        host = request.match_info.get("host", "localhost")
        port = int(request.match_info.get("port", "1234"))
        path = request.match_info.get("path", "ws")
        target_url = f"ws://{host}:{port}/{path}"
        logger.info("=> %s", target_url)

        if target_url in self._forward_map:
            raise web.HTTPForbidden()

        forwarder = ForwardConnection(request, target_url)
        self._forward_map[target_url] = forwarder
        await forwarder.connect()
        self._forward_map.pop(target_url)

    # -----------------------------
    # relay server infrastructure
    # -----------------------------

    async def relay_connect(self, request):
        id = request.path

        if id not in self._relay_map:
            handler = SinkConnection(request)
            self._relay_map[id] = handler

        handler = self._relay_map[id]
        if not handler.can_handle(request):
            raise web.HTTPForbidden()

        if await handler.connect(request):
            # Only pop when the server dies
            self._relay_map.pop(id)


# -----------------------------------------------------------------------------
# Executable
# -----------------------------------------------------------------------------


def main(host=None, port=None, www_path=None, proxy_route=None, mode=None):
    wsRelay = WsHandler()

    # Handle CLI
    parser = argparse.ArgumentParser(
        description="Start ws relay with static content delivery",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--host",
        type=str,
        default="localhost",
        help="the interface for the web-server to listen on (default: 0.0.0.0)",
        dest="host",
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=8080,
        help="port number for the web-server to listen on (default: 8080)",
        dest="port",
    )
    parser.add_argument(
        "--mode",
        type=str,
        default="forward",
        help="Working mode [forward, relay] (default: forward)",
    )
    parser.add_argument("--www", type=str, help="Directory to serve", dest="www_path")
    parser.add_argument(
        "--proxy-route",
        type=str,
        help="Proxy URL pattern (default: /proxy/{port}) mode::forward(ws://{host=localhost}:{port=1234}/{path=ws})",
        default="/proxy/{port}",
        dest="proxy_route",
    )
    args, _ = parser.parse_known_args()

    if host is None:
        host = args.host

    if port is None:
        port = args.port

    if mode is None:
        mode = args.mode

    if www_path is None:
        www_path = args.www_path

    if proxy_route is None:
        proxy_route = args.proxy_route

    # Manage routes
    routes = []

    # Need to be first: static delivery should be a fallback
    if proxy_route is not None:
        logger.info("Proxy route: %s", proxy_route)
        routes.append(web.get(proxy_route, wsRelay.get_handler(mode)))

    # Serve static content
    if www_path is not None:
        logger.info("WWW: %s", www_path)
        routes.append(web.get("/", _root_handler))
        routes.append(web.static("/", www_path))

    # Setup web app
    logger.info("Starting relay server: %s %s", host, port)
    web_app = web.Application()
    web_app.add_routes(routes)
    web.run_app(web_app, host=host, port=port)


# -----------------------------------------------------------------------------
# Main
# -----------------------------------------------------------------------------

if __name__ == "__main__":
    main()



import asyncio
import logging
import uuid
from pathlib import Path
import shutil

from wslink.protocol import WslinkHandler, AbstractWebApp


class WsConnection:
    def __init__(self):
        self._id = str(uuid.uuid4()).replace("-", "")
        self._ws = None
        self._on_message_fn = None
        self.closed = True

    # -------------------------------------------------------------------------
    # Method to be used by user
    # -------------------------------------------------------------------------

    def on_message(self, callback):
        self._on_message_fn = callback

    async def send(self, is_binary, msg):
        await self._ws.onMessage(is_binary, msg, self.client_id)

    async def close(self):
        await self.on_close(self._ws)

    # -------------------------------------------------------------------------
    # Method used by FakeWS
    # -------------------------------------------------------------------------

    @property
    def client_id(self):
        return self._id

    async def on_connect(self, ws):
        self.closed = False
        self._ws = ws
        await self._ws.onConnect(
            {"type": "generic", "connection": self}, self.client_id
        )

    async def on_close(self, ws):
        self.closed = True
        if self._ws == ws:
            ws.disconnect(self.client_id)
            self._ws = None

    async def send_str(self, value):
        await self._on_message_fn(False, value)

    async def send_bytes(self, value):
        await self._on_message_fn(True, value)


class WsEndpoint(WslinkHandler):
    def __init__(self, protocol=None, web_app=None):
        super().__init__(protocol, web_app)

    async def connect(self):
        conn = WsConnection()
        self.connections[conn.client_id] = conn
        await conn.on_connect(self)
        return conn

    async def disconnect(self, client_or_id):
        client_or_id = (
            client_or_id if isinstance(client_or_id, str) else client_or_id.client_id
        )
        if client_or_id in self.connections:
            client = self.connections.pop(client_or_id)
            await client.onClose(client_or_id)


class GenericServer(AbstractWebApp):
    def __init__(self, server_config):
        AbstractWebApp.__init__(self, server_config)
        self._websockets = {}
        self._stop_event = asyncio.Event()

        if "ws" in server_config:
            for route, server_protocol in server_config["ws"].items():
                protocol_handler = WsEndpoint(server_protocol, self)
                self._websockets[route] = protocol_handler

    def write_static_content(self, dest_directory):
        dest = Path(dest_directory)
        dest.mkdir(exist_ok=True, parents=True)
        if "static" in self._config:
            static_routes = self._config["static"]
            for route in sorted(static_routes.keys()):
                server_path = static_routes[route]
                if route == "/":
                    src = Path(server_path)
                    for child in src.iterdir():
                        if child.is_dir():
                            shutil.copytree(child, dest / child.name)
                        else:
                            shutil.copy2(child, dest / child.name)
                else:
                    shutil.copytree(server_path, dest / route)

    def __getattr__(self, attr):
        return self._websockets.get(attr, None)

    def __getitem__(self, name):
        return self._websockets.get(name, None)

    @property
    def ws_endpoints(self):
        return list(self._websockets.keys())

    async def start(self, port_callback):
        if port_callback is not None:
            port_callback(self.get_port())

        self._stop_event.clear()
        await self._stop_event.wait()

    async def stop(self):
        self._stop_event.set()


def startWebServer(*args, **kwargs):
    raise NotImplementedError("Generic backend does not provide a launcher")


def create_webserver(server_config):
    if "logging_level" in server_config and server_config["logging_level"]:
        logging.getLogger("wslink").setLevel(server_config["logging_level"])

    # Reverse connection
    if "reverse_url" in server_config:
        raise NotImplementedError("Generic backend does not support reverse_url")

    return GenericServer(server_config)



import asyncio
import os
import logging
import sys
import uuid

from wslink.protocol import WslinkHandler, AbstractWebApp

# Backend specific imports
import aiohttp
import aiohttp.web as aiohttp_web


# 4MB is the default inside aiohttp
MAX_MSG_SIZE = int(os.environ.get("WSLINK_MAX_MSG_SIZE", 4194304))
HEART_BEAT = int(os.environ.get("WSLINK_HEART_BEAT", 30))  # 30 seconds

logger = logging.getLogger(__name__)

# -----------------------------------------------------------------------------
# HTTP helpers
# -----------------------------------------------------------------------------


async def _root_handler(request):
    if request.query_string:
        return aiohttp.web.HTTPFound(f"index.html?{request.query_string}")
    return aiohttp.web.HTTPFound("index.html")


def _fix_path(path):
    if not path.startswith("/"):
        return "/{0}".format(path)
    return path


# -----------------------------------------------------------------------------


class WebAppServer(AbstractWebApp):
    def __init__(self, server_config):
        AbstractWebApp.__init__(self, server_config)
        self.set_app(aiohttp_web.Application())
        self._ws_handlers = []
        self._site = None
        self._runner = None

        if "ws" in server_config:
            routes = []
            for route, server_protocol in server_config["ws"].items():
                protocol_handler = AioHttpWsHandler(server_protocol, self)
                self._ws_handlers.append(protocol_handler)
                routes.append(
                    aiohttp_web.get(_fix_path(route), protocol_handler.handleWsRequest)
                )

            self.app.add_routes(routes)

        if "static" in server_config:
            static_routes = server_config["static"]
            routes = []

            # Ensure longer path are registered first
            for route in sorted(static_routes.keys(), reverse=True):
                server_path = static_routes[route]
                routes.append(
                    aiohttp_web.static(
                        _fix_path(route), server_path, append_version=True
                    )
                )

            # Resolve / => index.html
            self.app.router.add_route("GET", "/", _root_handler)
            self.app.add_routes(routes)

        self.app["state"] = {}

    # -------------------------------------------------------------------------
    # Server status
    # -------------------------------------------------------------------------

    @property
    def runner(self):
        return self._runner

    @property
    def site(self):
        return self._site

    def get_port(self):
        """Return the actual port used by the server"""
        return self.runner.addresses[0][1]

    # -------------------------------------------------------------------------
    # Life cycles
    # -------------------------------------------------------------------------

    async def start(self, port_callback=None):
        self._runner = aiohttp_web.AppRunner(
            self.app, handle_signals=self.handle_signals
        )

        logger.info("awaiting runner setup")
        await self._runner.setup()

        self._site = aiohttp_web.TCPSite(
            self._runner, self.host, self.port, ssl_context=self.ssl_context
        )

        logger.info("awaiting site startup")
        await self._site.start()

        if port_callback is not None:
            port_callback(self.get_port())

        logger.info("Print WSLINK_READY_MSG")
        STARTUP_MSG = os.environ.get("WSLINK_READY_MSG", "wslink: Starting factory")
        if STARTUP_MSG:
            # Emit an expected log message so launcher.py knows we've started up.
            print(STARTUP_MSG)
            # We've seen some issues with stdout buffering - be conservative.
            sys.stdout.flush()

        logger.info(f"Schedule auto shutdown with timout {self.timeout}")
        self.shutdown_schedule()

        logger.info("awaiting running future")
        await self.completion

    async def stop(self):
        # Disconnecting any connected clients of handler(s)
        for handler in self._ws_handlers:
            await handler.disconnectClients()

        # Neither site.stop() nor runner.cleanup() actually stop the server
        # as documented, but at least runner.cleanup() results in the
        # "on_shutdown" signal getting sent.
        logger.info("Performing runner.cleanup()")
        await self.runner.cleanup()

        # So to actually stop the server, the workaround is just to resolve
        # the future we awaited in the start method.
        logger.info("Stopping server")
        self.completion.set_result(True)


class ReverseWebAppServer(AbstractWebApp):
    def __init__(self, server_config):
        super().__init__(server_config)
        self._url = server_config.get("reverse_url")
        self._server_protocol = server_config.get("ws_protocol")
        self._ws_handler = AioHttpWsHandler(self._server_protocol, self)

    async def start(self, port_callback=None):
        if port_callback is not None:
            port_callback(0)

        await self._ws_handler.reverse_connect_to(self._url)

    async def stop(self):
        client_id = self._ws_handler.reverse_connection_client_id
        ws = self._ws_handler.connections[client_id]
        await ws.close()


def create_webserver(server_config):
    if "logging_level" in server_config and server_config["logging_level"]:
        logging.getLogger("wslink").setLevel(server_config["logging_level"])

    # Shortcut for reverse connection
    if "reverse_url" in server_config:
        return ReverseWebAppServer(server_config)

    # Normal web server
    return WebAppServer(server_config)


# -----------------------------------------------------------------------------
# WS protocol definition
# -----------------------------------------------------------------------------


def is_binary(msg):
    return msg.type == aiohttp.WSMsgType.BINARY


class AioHttpWsHandler(WslinkHandler):
    async def disconnectClients(self):
        logger.info("Closing client connections:")
        keys = list(self.connections.keys())
        for client_id in keys:
            logger.info("  {0}".format(client_id))
            ws = self.connections[client_id]
            await ws.close(
                code=aiohttp.WSCloseCode.GOING_AWAY, message="Server shutdown"
            )

        self.publishManager.unregisterProtocol(self)

    async def handleWsRequest(self, request):
        client_id = str(uuid.uuid4()).replace("-", "")
        current_ws = aiohttp_web.WebSocketResponse(
            max_msg_size=MAX_MSG_SIZE, heartbeat=HEART_BEAT
        )
        self.connections[client_id] = current_ws

        logger.info("client {0} connected".format(client_id))

        self.web_app.shutdown_cancel()

        try:
            await current_ws.prepare(request)
            await self.onConnect(request, client_id)
            async for msg in current_ws:
                await self.onMessage(is_binary(msg), msg, client_id)
        finally:
            await self.onClose(client_id)

            del self.connections[client_id]
            self.authentified_client_ids.discard(client_id)

            logger.info("client {0} disconnected".format(client_id))

            if not self.connections:
                logger.info("No more connections, scheduling shutdown")
                self.web_app.shutdown_schedule()

        return current_ws

    async def reverse_connect_to(self, url):
        logger.debug("reverse_connect_to: running with url %s", url)
        client_id = self.reverse_connection_client_id
        async with aiohttp.ClientSession() as session:
            logger.debug("reverse_connect_to: client session started")
            async with session.ws_connect(url) as current_ws:
                logger.debug("reverse_connect_to: ws started")
                self.connections[client_id] = current_ws
                logger.debug("reverse_connect_to: onConnect")
                await self.onConnect(url, client_id)

                async for msg in current_ws:
                    if not current_ws.closed:
                        await self.onMessage(is_binary(msg), msg, client_id)

                logger.debug("reverse_connect_to: onClose")
                await self.onClose(client_id)
                del self.connections[client_id]

        logger.debug("reverse_connect_to: exited")
